#ifdef __OPTIMIZE__
#define NO_MSGS
#endif

#include <math.h>
#include <stdlib.h>

#include "Msgs.h"
#include "physics.h"
#include "traverse.h"
#include "tree.h"
#include "stk.h"
#include "vop.h"
#include "do_grav.h"
#include "integrate.h"
#include "placebods.h"

/* rcrit_fac shows up in the MAC, where the test is:
   if ( cp->rcrit > rcritfac * sink->bmax)
     SplitSrc
   else
     SplitSink

   This test is only applied after the 'guaranteed' tests have come up
     undecided.  I.e., 
     - when the sink is entirely inside the rcrit ball, we ALWAYS
     split the source,
     - when the rcrit ball is entirely inside the sink we ALWAYS split the sink
     - when the entire sink is outside rcrit, we ALWAYS accept.

   The rcritfac desision is made only when there is some incomplete, but
   non-empty overlap between the rcrit ball, and the sink.

   Setting rcritfac to 0 splits sources whenever the rcrit ball and the
     sink partially overlap.  This forces the source traversal to continue
     as far as it can with the current sink, eventually identifying all
     possible interactions that are acceptable for the entire sink.
     This is the most "Laughable" situation, but it doesn't necessarily
     result in the fewest possible MACs.  It does result in more interactions.
   Setting rcritfac to infinity splits sinks at every opportunity - i.e.,
     if we aren't certain that a source is acceptable for a given sink,
     we just defer it till we get a look at a smaller sink.  This
     results in a pure BH strategy.  A faster way to achieve the same 
     thing is to set max_sinksplit_per_body=0 (see below).
   Intermediate values of rcritfac have the following properties:
     increasing rcritfac => decreasing #interactions
     decreasing rcritfac => increasing #interactions

   The number of MACs is not monotonic in rcrit_fac.  Should that be telling
     us something?

   Since we get the most Mflops from interactions, a low value of rcritfac
     leads to the most Mflops.  Conventional wisdom from relerr was that
     rcritfac=4.0 was the best value, but rcritfac=2.0 gave more Mflops.
*/
extern float rcritfac;

/* max_sinksplit_per_body governs how many times we are allowed to
   split sinks.  After finding a 'common' ancestor, the DL traversal will
   split sinks max_sinksplit_per_body times.  Setting this to 0 results
   in a pure BH strategy.  The default (1) is set in cmdline.c.  */
extern int max_sinksplit_per_body;

/* Counters and timers. */
Counter_t DLMacCnt, DLUncertainMacCnt, DLUncertainSplitSinkCnt;
Counter_t DLPeelCnt, DLDepthCnt;

Timer_t DLPollWaste;

/* file-local variables used by callbacks */
static const float *BodyPos;		/* position of body - read only */
static float MassAccum;
static float eps2;
static Ivec_t *iv, baseivec[16384];

static float GNewt;
static gravdiags_t *diags;

static void InteractWithSg(oocptr sgooc);
static int InteractWithTc(const tcell *);

typedef struct {
  /* Used when popping back levels, so we know how to reset iv */
  Ivec_t *firstiv;
  /* Used for diagnostics */
  int daughtercnt;
  /* Used for error bounds and evaluation */
  float pos[NDIM];
  float bmax;
  /* We could manage with only one rejects and locks Stk, much like
     the single baseivec, but it would require even more bookkeeping */
  Stk rejects;
  Stk locks;
} Sink;

static Sink SinkTbl[64];	/* deep enough for any reasonable tree. */
static int sinklev;

/* Copy the tbodies in an sg to 'ivec' */
static void
DLDoBodies(const sib_group *sg){
  tbody *bp = Tbodies(sg);
  int nbod = sg->ntbody;
  int i;

  Msg("DL", ("DLDoBodies: sg=%s\n", PrintSg(sg)));
  for(i=0; i<nbod; i++){
    iv->u.used.mass = bp->mass;
    VV(iv->u.used.pos, = bp->pos);
    SinkTbl[sinklev].daughtercnt++;
    bp++;
    iv++;
  }
}

/* Split the sink by putting it on the reject stack at the current
   sinklev.  When we return from the enclosing MAC, the
   traversal/paging code may decide to flush the copy of the tc that
   we are currently looking at.  Therefore, we need to store an oocptr
   that can be re-instantiated when we come back to look at this src
   cell with a new sink.  They go into the locks Stk. Alternatively,
   we could Lock this page into memory, and be careful to unlock it
   when we pop back up the Sink stack.  */
static Stk *rejects;
static Stk *locks;

#define SPLIT_SRC 10
#define SPLIT_SINK 11

/* The Don't Laugh mac.  If it 'likes' an interaction, it saves it up
   for later.  Failures may either be SPLIT_SRC (i.e., Split the sources,
   return MAC_REJECT), or the SPLIT_SINK (i.e., ask me about this source again
   at the next sinklevel). */
static int
DLGravMac(const tcell *tc){
  Sink *sink = &SinkTbl[sinklev];
  cell *cp = Contents(tc);
  float r[NDIM], r2, rcritplus, rcritminus;

  IncrCounter(&DLMacCnt);
  Msg("DL", ("Mac(tc=%s, sinklev=%d, sinkbmax=%g)\n", 
	     PrintTc(tc), sinklev, sink->bmax));
  Msg("DL", ("Contents(tc): %s\n", PrintCellContents(cp)));

  VVV(r, = sink->pos, - cp->pos);
  r2 = Dot(r, r);
  rcritplus = cp->rcrit + sink->bmax;
  rcritminus = cp->rcrit - sink->bmax;
  /* sink->bmax==0 should short-circuit this so the only possibilities
     are AcceptSrc or SplitSrc.  if sink->bmax is 0, we should never
     continue to split the sink.  

     Proof: If sink->bmax==0, then rcritplus==rcritminus==rcrit.
     Therefore we always get into the outer if() or else if() clause.
     Never the final else.  If we get into elseif, then we must have
     rcrit>=0 (otherwise we would have had r2>=rcritplus*rcritplus), so
     we SplitSrc.  The only possibility in the if() clause is Accept, so
     the code here is 'correct' (as of May 19, 1998) */
  if( r2 >= rcritplus*rcritplus ){
    /* This cell is sufficiently far for EVERYTHING in the sink */
    Msg("DL", ("r2>rcritplus^2, tcbits=%d, rcritplus = %g, r=%g MAC_ACCEPT\n",
	       tc->bits, rcritplus, sqrt(r2)));
    return MAC_ACCEPT;
  }else if(r2 < rcritminus*rcritminus){
    Msg("DL", ("r2<rcritminus^2, tcbits=%d, rcritminus = %g, r=%g...", 
		 tc->bits, rcritminus, sqrt(r2)));
    if( rcritminus > 0 ){
      /* r < (rcrit-bmax) */
      /* The entire sink is completely inside the source's rejection
	 zone.  Nothing inside the sink can interact with this source.Split
	 the source */
      Msg("DL", ("rcritminus>0, SPLIT_SRC\n"));
      return SPLIT_SRC;
    }else{
      /* r < (bmax - rcrit) */
      /* The entire source rejection-zone is completely inside the sink.
	 There are certainly some sinks that can't interact with this
	 source */
      Msg("DL", ("rcritminus<=0, SPLIT_SINK\n"));
      return SPLIT_SINK;
    }
  }else{ 
    /* At this point, it's not clear-cut.  We need a better theory.
       Comparing cp->rcrit with sink->bmax makes sense if we believe
       that cutting the larger one in "half" will increase the
       MAC_ACCEPTable volume of space by the largest amount.  Another
       interesting choice is to always do SPLIT_SINK (i.e., 
       rcritfac=infinity) , which results in exactly the same number
       of interactions as the 'NlgN' method.  Whenever we choose to do
       a SPLIT_SRC here, we risk doing a few extra interactions (when
       that source would have been acceptable to some of the bodies in
       the sink), but we may save MAC's. */
    IncrCounter(&DLUncertainMacCnt);
    if( cp->rcrit > rcritfac * sink->bmax ){
      Msg("DL", ("rcrit (%g)> rcritfac * bmax(%g) SPLIT_SRC\n",
		 cp->rcrit, sink->bmax));
      return SPLIT_SRC;
    }else{
      Msg("DL", ("rcrit (%g) <= rcritfac * bmax(%g) SPLIT_SINK\n",
		 cp->rcrit, sink->bmax));
      IncrCounter(&DLUncertainSplitSinkCnt);
      return SPLIT_SINK;
    }
  }
}

/* Are the 'rejects' in SinkTbl[lev] usable by the new body, i.e., is
   this body contained in lev-1. */
/* Would a key comparison be better? */
static int
SinkUsableP(int lev){
  Vxd(float dist);
  Sink *s;

  if( lev == 0 )
    return 1;

  s = &SinkTbl[lev-1];
  VxVV(dist, = s->pos, - BodyPos);
  return s->bmax*s->bmax > Dotx(dist, dist);
}

/* This might be slightly better if it didn't actually issue the
   recursive call to InteractWithSg.  If the recursion is deferred
   till some unspecified later time, there >>might<< be better cache
   re-use since InteracWithTc is usually called on consecutive, or
   at least nearby-in-memory locations.  In order to defer the recursion,
   the caller would have to check for a return value of, e.g., SPLIT_SRC,
   and remember to call InteractWithSg later. */
static int
InteractWithTc(const tcell *tc){
    int mac;
    Ivec_t *interact;
    cell *cp;

    mac = DLGravMac(tc);
    switch(mac){
    case SPLIT_SRC:
      InteractWithSg(tc->daughter);
      break;
    case SPLIT_SINK:
      Msg("DL", ("pushing tc onto reject stack: %s\n", PrintTc(tc)));
      StkPushData(rejects, &tc, sizeof(tc));
      break;
    case MAC_ACCEPT:
      interact = iv++;
      cp = Contents(tc);
      Msg("DL", ("ivec[%ld]: cell at (" Sinfix("%g", " ") "), daughters=%d bits=%d\n",
		 (long)(iv-baseivec),
		 Vinfix(cp->pos, COMMA), cp->daughters,
		 tc->bits));
      interact->u.used.mass = cp->mass;
      VV(interact->u.used.pos, = cp->pos);
      SinkTbl[sinklev].daughtercnt += cp->daughters;
      break;
    }
    return mac;
}

static void
InteractWithSg(oocptr sgooc){
  sib_group *sg;
  PgHndl sghndl;
  int i;
  int nsplitsink = 0;

  sg = OOCFollow(sgooc, &sghndl);
  DLDoBodies(sg);
  for(i=0; i<sg->ncell; i++){
    tcell *tc;

    tc = &sg->tcells[i];
    if( InteractWithTc(tc) == SPLIT_SINK )
      nsplitsink++;
  }

  if( nsplitsink > 0 ){
    StkPushData(locks, &sghndl, sizeof(sghndl));
  }else{
    PgUnref(sghndl, PG_CLEAN);
  }
}

static void
DLEachBody(const body *bp_orig){
  int ncut;
  body copy = *bp_orig;
  body *bp = &copy;
  int max_sinklev;
  int pollcnt = 0;		/* a hack to control how often we Poll... */
  int poll_freq = 2;

  /* Set up appropriate static vars so the callbacks know where to
	 find things */
  BodyPos = bp->pos;
  Msg("DL", ("DL Begin new body key=%s, at (" Sinfix("%g", " ") ")\n",
	     PrintKey(bp->key), Vinfix(bp->pos, COMMA)));
  while( !SinkUsableP(sinklev) ){
    /* peel back by unlocking all the pages locked by this level */
    PgHndl *hptr, *end;
    hptr = StkBase(&SinkTbl[sinklev].locks);
    end = (PgHndl*)( (char *)hptr + StkSz(&SinkTbl[sinklev].locks) );
    Msg("DL", ("peel back sinklev, Unref %ld locks at level %d\n", 
	       (long)(hptr-end), sinklev));
    while( hptr < end ){
      PgUnref(*hptr, PG_CLEAN);
      hptr++;
    }
    StkClear(&SinkTbl[sinklev].locks);
    StkClear(&SinkTbl[sinklev].rejects);
    --sinklev;
    assert(sinklev >= 0);
    IncrCounter(&DLPeelCnt);
  }
  max_sinklev = sinklev + max_sinksplit_per_body;

  rejects = &SinkTbl[sinklev].rejects;
  iv = SinkTbl[sinklev].firstiv;
  --sinklev;

  while( StkSz(rejects) ) {
    const tcell **tc, **lasttc;
    tbbox cbb;
    Key_t bkey;

    sinklev++;
    assert(sinklev >= 0 );
    SinkTbl[sinklev].firstiv = iv;
    /* This used to be called 'Inherit' */
    assert( sinklev <= max_sinklev );
    if( sinklev == max_sinklev ){
      VV(SinkTbl[sinklev].pos, = BodyPos);
      SinkTbl[sinklev].bmax = 0.F;
    }else{
      float sz[NDIM];
      bkey = KeyRshift(bp->key, (nbodybits - sinklev)*NDIM);
      CellBBFromKey(bkey, &UniverseBbox, &cbb, key_order);
      CenterBbox(&cbb, SinkTbl[sinklev].pos, sz);
      SinkTbl[sinklev].bmax = sqrt(Dot(sz, sz)) * 0.5;
    }
    Msg("DL", ("Dropping down to sinklev=%d, bkey=%s, center=(" Sinfix("%g", " ") ")\n",
	       sinklev, PrintKey(bkey), Vinfix(SinkTbl[sinklev].pos, COMMA)));

    tc = StkBase(rejects);
    lasttc = (const tcell **)( (char *)tc + StkSz(rejects) );

    rejects = &SinkTbl[sinklev+1].rejects;
    locks = &SinkTbl[sinklev+1].locks;
    assert(StkSz(rejects)==0);
    assert(StkSz(locks)==0);

    while( tc < lasttc ){
      InteractWithTc(*tc++);
    }
    if( ++pollcnt >= poll_freq ){
      StartTimer(&DLPollWaste);
      PagePoll();
      StopTimer(&DLPollWaste);
      pollcnt=0;
    }
  }

  /* Zero the accumulators and actually do the interactions*/
  MassAccum = 0.F;
  VS(bp->acc, = 0.F);
  bp->phi = 0.F;
  ncut = 0;
  do_grav(baseivec, iv, bp->pos, &MassAccum, bp->acc, &bp->phi, &eps2, &ncut);
  bp->nterms = iv - baseivec;
  VS(bp->acc, *= GNewt);
  bp->phi *= GNewt;
  Msgf(("body %d Nterms=%d, acc=(%g %g %g)\n", bp->ident, bp->nterms,
	bp->acc[0], bp->acc[1], bp->acc[2]));
  AddCounter(&DLDepthCnt, sinklev);

  /* Push the body to its new position and velocity. */
  DiagsAdd(diags, bp, 0.5*dt);
  IntegrateBody(bp);
  bp->key = KeyFromFloats(bp->pos, &UniverseBbox, nbodybits, key_order);
  if( KeyEQZ(bp->key) ){
    Error("New position (%g %g %g) out of UniverseBbox\n",
	  bp->pos[0], bp->pos[1], bp->pos[2]);
  }
  /* Save it into the appropriate bucket.  If buffers fill up,
	 this will generate communication. */
  PlaceBody(bp, bp->key);
}

void
DLTraverse(float eps, float GNewt_arg, gravdiags_t *diags_arg){
  int b;
  bucket_t *tmp;
  int lev;
  float UniSz[NDIM];

  GNewt = GNewt_arg;
  diags = diags_arg;
  assert(nbodybits <= sizeof(SinkTbl)/sizeof(*SinkTbl) );
  eps2 = eps*eps;
  PlaceBodsBegin(NDIM, sizeof(body), decompranges, nbuckets, nextbodies, bucketdividers);
  DiagsZero(diags);

  /* Initialize loop-variables/state */
  for(lev=0; lev < nbodybits; lev++){
    StkInitEz(&SinkTbl[lev].locks);
    StkInitEz(&SinkTbl[lev].rejects);
  }

  iv = baseivec;
  sinklev = 0;
  /* Put the whole (sink) Universe into SinkTbl[0].  Traverse it once
     to get a list of 'rejects', that correspond to source cells that
     returned with SPLIT_SINK against the sink Universe.  This is not
     necessarily just the root of the tree.  It's reasonable (likely?)
     that the source root and its immediate descendants returned SPLIT_SRC
     which means we never have to look at them again - we start with
     their descendants when we look at subsequent sink bodies.
  */
  CenterBbox(&UniverseBbox, SinkTbl[0].pos, UniSz);
  SinkTbl[0].bmax = sqrt( Dot( UniSz, UniSz ) )* .5001 ;
  rejects = &SinkTbl[0].rejects;
  locks = &SinkTbl[0].locks;
  InteractWithSg(srcroot);
  SinkTbl[0].firstiv = iv;
  Msg("DL", ("After traversing Universe, %ld interactions, %ld top-level tcells\n",
	     (long)(iv-baseivec), (long)(StkSz(rejects)/sizeof(tcell *))));

  /* Call DLEachBody for every body.  This is where the real work is. */
  doBodies(DLEachBody);

  /* We are now done with all the bodies.  They are safely updated and
     stored in newbodies */
  for(b=0; b<nbuckets; b++){
    BucketDestroy(&bodies[b]);
  }

  PagePollTillDone();
  PlaceBodsEnd();
  DiagsFinish(diags);

  /* Final cleanup/peelback. */
  while( sinklev>=0 ){
    PgHndl *hptr, *end;
    hptr = StkBase(&SinkTbl[sinklev].locks);
    end = (PgHndl*)( (char *)hptr + StkSz(&SinkTbl[sinklev].locks) );
    while( hptr < end ){
      PgUnref(*hptr, PG_CLEAN);
      hptr++;
    }
    StkClear(&SinkTbl[sinklev].locks);
    StkClear(&SinkTbl[sinklev].rejects);
    --sinklev;
    IncrCounter(&DLPeelCnt);
  }

  for(lev=0; lev < nbodybits; lev++){
    StkTerminate(&SinkTbl[lev].locks);
    StkTerminate(&SinkTbl[lev].rejects);
  }

  /* Swap nextbodies and bodies */
  tmp = nextbodies;
  nextbodies = bodies;
  bodies = tmp;
}
